// Clock software for SHE879 six-digit clock kit sold very cheap on eBay, aliexpress, banggood etc. Complete rewrite
// Target: SHE879 with Atmel AT89C2051 (MCS51 MCU with 2k FLASH) using 12 MHz XTAL.
// Compiled with SDCC Compiler Version 3.9.0 using: sdcc -mmcs51 -pedantic --model-small --std-c99 she879.c
// tcfkat 20190618

#ifndef __SDCC_mcs51
#error "Please use SDCC with -mmcs51"
#endif

#define MICROCONTROLLER_AT89CX051
#include <mcs51reg.h>
#include <stdint.h>
#include <stdbool.h>

#define BUTTON             P3_7

// Do not change!
#define NMBRDIG            6
enum DIGITS   { SECONES = 0, SECTENS, MINONES, MINTENS, HOURONES, HOURTENS };
enum MODES    { RUN = 0, SET };
enum SETMODES { MINUTES = 0, HOURS, DONE };
enum KEYPRESS { NONE = 0, SHORT, LONG };

// Global data but local to this module. Initialized to 0 by startup code
__data static volatile uint8_t  disp[NMBRDIG];  // display buffer
__data static volatile uint8_t  ctr4ms;         // 4 millisecond counter
__data static volatile _Bool    sync;           // one second elapsed, increment clock
__data static volatile _Bool    key;            // true = button pressed
__data static uint8_t           mode;           // RUN or SET

// On SHE879, the segments are scrambled on Port1: 0 = a, 1 = b, 2 = f, 3 = g, 4 = d, 5 = e, 6 = c
__code static const uint8_t seg[11]      = {0x88,0xBD,0xC4,0xA4,0xB1,0xA2,   // '0','1','2','3','4','5',
                                            0x82,0xBC,0x80,0xA0,0xFF};       // '6','7','8','9',' '
__code static const uint8_t dig[NMBRDIG] = {0xDF,0xF7,0xEF,0xFB,0xFD,0xFE};

// Protos
static void  external0isr_unused (void) __interrupt (0) __using(1);
static void  timer0isr_250us     (void) __interrupt (1) __using(1);
static void  external1isr_unused (void) __interrupt (2) __using(1);
static void  timer1isr_unused    (void) __interrupt (3) __using(1);
static _Bool inc_seconds         (void);
static _Bool inc_minutes         (void);                            // Used in set and run mode
static void  inc_hours           (void);                            // Used in set and run mode
static void  set_clk             (void);                            // Manually set clock

void main(void){
    EA = 0;          // Disable all interrupts
    P1 = P3 = 0xFF;
    PCON = 0;
    TMOD = 0x02;     // Timer 0 Mode 2 (8 Bit Autoreload)
    TL0 = TH0 = 6;   // 250 microseconds timer delay
    IP = IE = 0;     // No special priority
    ET0 = 1;         // Enable interrupt for timer 0
    TCON = 0x10;     // Enable timer 0
    EA = 1;          // Enable interrupts
    disp[SECONES] = disp[SECTENS]  = disp[MINONES] =
    disp[MINTENS] = disp[HOURONES] = disp[HOURTENS] = 8;    // Lamptest: 88 88 88
    while (!sync);                                          // Wait one second
    disp[SECONES] = disp[SECTENS]  = disp[MINONES] =
    disp[MINTENS] = disp[HOURONES] = disp[HOURTENS] = 0;    // Display 00 00 00
    sync = false;
    mode = SET;      // Force clock setting at startup
    // Main loop
    while (true){
        if (key)
            mode = SET;
        switch (mode){
            case RUN:
                if (sync){
                    if (inc_seconds())
                        if (inc_minutes())
                            inc_hours();
                    sync = false;
                }
                break;
            case SET:
                set_clk();
                __critical{      // disable interrupts
                    ctr4ms = 0;
                    sync = false;
                }
                mode = RUN;
                break;
            default:
                break;
        }
    }
} // main()

static void set_clk (void)
{
__data uint8_t tmp1, tmp2, smode, keyp;

    disp[SECONES] = disp[SECTENS] = 0; // Reset seconds to 00
    while (key);                       // Wait for button release
    smode = MINUTES;
    tmp1 = disp[MINONES];
    tmp2 = disp[MINTENS];
    while (smode != DONE){
        keyp = NONE; // key has not been pressed
        if (key){
            __critical{         // disable interrupts
                ctr4ms = 0;
                sync = false;
            }
            while (key);        // Wait for button release
            if (ctr4ms > 50 || sync)
                keyp = LONG;    // key has been pressed for longer than 200ms
            else
                keyp = SHORT;   // key has been pressed for less than 200ms
        }
        // blink two digits that are to be set
        if (ctr4ms < 125){
            disp[smode * 2 + 2] = tmp1;
            disp[smode * 2 + 3] = tmp2;
        }else{
            disp[smode * 2 + 2] = disp[smode * 2 + 3] = 10; // Blank two digits
        }
        if (keyp == SHORT){
            disp[smode * 2 + 2] = tmp1;  // Restore in case display was blanked in
            disp[smode * 2 + 3] = tmp2;  // moment when button was short pressed
            if (smode == MINUTES)
                inc_minutes();
            else if (smode == HOURS)
                inc_hours();
            tmp1 = disp[smode * 2 + 2];
            tmp2 = disp[smode * 2 + 3];
        }else if (keyp == LONG){
            disp[smode * 2 + 2] = tmp1;  // Restore in case display was blanked in
            disp[smode * 2 + 3] = tmp2;  // moment when button was long pressed
            smode ++;
            tmp1 = disp[smode * 2 + 2];  // Get actual value of next digits
            tmp2 = disp[smode * 2 + 3];  // if smode == DONE read beyond disp[], but no problem
        }
    }
}

static _Bool inc_seconds (void)
{
    disp[SECONES]++;
    if (disp[SECONES] > 9){
        disp[SECONES] = 0;
        disp[SECTENS]++;
        if (disp[SECTENS] > 5){
            disp[SECTENS] = 0;
            return true;       // Increment minutes also
        }
    }
    return false;
}

static _Bool inc_minutes (void)
{
    disp[MINONES]++;
    if (disp[MINONES] > 9){
        disp[MINONES] = 0;
        disp[MINTENS]++;
        if (disp[MINTENS] > 5){
            disp[MINTENS] = 0;
            return true;       // Increment hours also
        }
    }
    return false;
}

static void inc_hours (void)
{
    disp[HOURONES]++;
    if (disp[HOURONES] > 9 && disp[HOURTENS] < 2){
        disp[HOURONES] = 0;
        disp[HOURTENS]++;
    }else if (disp[HOURONES] > 3 && disp[HOURTENS] >= 2){
        disp[HOURONES] = disp[HOURTENS] = 0;
    }
}

// Timer 0 is in Autoreload mode, because of the stupid MCS51 timer design this is the only mode that ensures
// that no counts are lost until the ISR is entered and counter is reloaded manually within the ISR. As there is
// no prescaler on MCS51 this ISR is called every 250 machine cycles (= 250 microseconds w/ 12 MHz XTAL) and should
// therefore be kept as short as possible. Using bank 1 helps by reducing the number of registers to be saved.
static void timer0isr_250us (void) __interrupt (1) __using(1)
{
__data static volatile uint8_t ctr250us  = 0;  // 250 microsecond counter
__data static volatile uint8_t actdig    = 0;  // actual displayed digit for multiplexing. 0 = rightmost

    ctr250us ++;
    if ( !(ctr250us & 0x07) ){
        // do LED multiplexing every 2 ms
        P3 = 0xFF; // turn all digits off
        actdig++;
        if (actdig >= NMBRDIG)
            actdig = 0;
        P1 = seg[disp[actdig]];
        P3 = dig[actdig];
        if ( !(ctr250us & 0x0F) ){
            // increment counter every 4ms
            ctr4ms ++;
            if (ctr4ms >= 250){
                ctr4ms = 0;
                sync = true;
            }
            key = !BUTTON;    // Scan button every 4ms, helps debouncing (hopefully)
        }
    }
}

// Stubs
static void external0isr_unused (void) __interrupt (0) __using(1){}
static void external1isr_unused (void) __interrupt (2) __using(1){}
static void timer1isr_unused    (void) __interrupt (3) __using(1){}
